<?php
/**
 * Bolt magento plugin
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Open Software License (OSL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/osl-3.0.php
 *
 * @category   Bolt
 * @package    Bolt_Boltpay
 * @copyright  Copyright (c) 2016-2020 Bolt Financial, Inc (https://www.bolt.com)
 * @license    http://opensource.org/licenses/osl-3.0.php  Open Software License (OSL 3.0)
 */

/**
 * Class Bolt_Boltpay_Model_Coupon
 *
 * Base Magento Bolt Coupon Model class
 *
 */
class Bolt_Boltpay_Model_Coupon extends Bolt_Boltpay_Model_Abstract
{
    const ERR_INSUFFICIENT_INFORMATION = 6200;
    const ERR_CODE_INVALID = 6201;
    const ERR_CODE_EXPIRED = 6202;
    const ERR_CODE_NOT_AVAILABLE = 6203;
    const ERR_CODE_LIMIT_REACHED = 6204;
    const ERR_MINIMUM_CART_AMOUNT_REQUIRED = 6205;
    const ERR_UNIQUE_EMAIL_REQUIRED = 6206;
    const ERR_ITEMS_NOT_ELIGIBLE = 6207;
    const ERR_SERVICE = 6001;

    /**
     * Response information sent to bolt
     */
    protected $responseStatus; // Bolt error status code
    protected $responseError; // Error description
    protected $responseCart; // Response quote
    protected $discountInfo; // Details about how much discount was applied as a result of coupon
    protected $httpCode; // HTTP response code

    protected $immutableQuote = null;
    protected $parentQuote = null; // Session quote
    protected $coupon = null;
    protected $rule = null;

    protected $requestObject = null;
    protected $mockTransaction = null;

    /**
     * Applies coupon from hook input to customer quote if the coupon is valid.
     *
     * @return bool Whether or not coupon was applied
     */
    public function applyCoupon()
    {
        try {

            if (!$this->requestObject){
                $msg = 'Need to set setup variables in order to apply coupon';
                $this->boltHelper()->logWarning($msg);
                throw new \Bolt_Boltpay_BadInputException($msg);
            }

            $this->validateCoupon();
            $this->applyCouponToQuotes();
            $this->validateAfterApplyingCoupon();

            $this->setSuccessResponse();
        } catch (\Bolt_Boltpay_BadInputException $e){
            return false;
        } catch (\Exception $e) {
            $this->boltHelper()->notifyException($e);
            $this->boltHelper()->logException($e);
            return false;
        }
        return true;
    }

    /**
     * Sets up any variables needed to process coupon.
     *
     * @param $requestObject
     */
    public function setupVariables($requestObject)
    {
        $this->requestObject = $requestObject;
        $this->mockTransaction = (object) array("order" => $requestObject );
    }

    /**
     * Makes several checks to confirm that coupon is valid.
     */
    protected function validateCoupon()
    {
        $this->validateEmptyCoupon();
        $this->validateCouponExists();
        $this->validateRuleExists();
        $this->validateCartIdentificationData();
        $this->validateOrderExists();
        $this->validateSessionQuote();
        $this->validateImmutableQuote();
        $this->validateEmptyCart();
        $this->validateToDateForRule();
        $this->validateFromDateForRule();
        $this->validateCouponUsageLimits();
        $this->validateCouponCustomerUsageLimits();
        $this->validateRuleCustomerUsageLimits();
    }

    /**
     * Verifies the coupon isn't an empty string.
     *
     * @throws Bolt_Boltpay_BadInputException if coupon code is empty
     */
    public function validateEmptyCoupon()
    {
        if ($this->getCouponCode() === '') {
            $this->setErrorResponseAndThrowException(
                self::ERR_CODE_INVALID,
                'No coupon code provided',
                422
            );
        }
    }

    /**
     * Verifies the coupon exists.
     *
     * @throws Exception
     */
    public function validateCouponExists()
    {
        $coupon = $this->getCoupon();
        if ($coupon->isEmpty() || $coupon->isObjectNew()) {
            $this->setErrorResponseAndThrowException(
                self::ERR_CODE_INVALID,
                sprintf('The coupon code %s was not found', $this->getCouponCode()),
                404
            );
        }
    }

    /**
     * Verifies the coupon rule exists.
     *
     * @throws Exception
     */
    public function validateRuleExists()
    {
        $rule = $this->getRule();
        if ($rule->isEmpty() || $rule->isObjectNew()) {
            $this->setErrorResponseAndThrowException(
                self::ERR_CODE_INVALID,
                sprintf('The coupon code %s was not found', $this->getCouponCode()),
                404
            );
        }
    }

    /**
     * Verifies that the identifying factors for the care parent quote, immutable quote, and increment id were passed
     * in.
     *
     * @throws Exception
     */
    public function validateCartIdentificationData()
    {
        $parentQuoteId = $this->getParentQuoteId();
        $incrementId = $this->boltHelper()->getIncrementIdFromTransaction($this->mockTransaction);
        $immutableQuoteId = $this->boltHelper()->getImmutableQuoteIdFromTransaction($this->mockTransaction);

        if (empty($parentQuoteId) || empty($incrementId) || empty($immutableQuoteId)) {
            $this->setErrorResponseAndThrowException(
                self::ERR_INSUFFICIENT_INFORMATION,
                'The order reference is invalid.',
                422
            );
        }
    }

    /**
     * Verifies that order doesn't already exist based on increment id.
     *
     * @throws Exception
     */
    public function validateOrderExists()
    {
        $incrementId = $this->boltHelper()->getIncrementIdFromTransaction($this->mockTransaction);
        /** @var Mage_Sales_Model_Order $order */
        $order = Mage::getModel('sales/order')->loadByIncrementId($incrementId);
        if ($order->getId()) {
            $this->setErrorResponseAndThrowException(
                self::ERR_INSUFFICIENT_INFORMATION,
                sprintf('The order #%s has already been created.', $incrementId),
                422
            );
        }
    }

    /**
     * Verifies that the session quote exists.
     *
     * @throws Exception
     */
    public function validateSessionQuote()
    {
        $parentQuote = $this->getParentQuote();

        if ($parentQuote->isEmpty() || !$parentQuote->getIsActive()) {

            $this->setErrorResponseAndThrowException(
                self::ERR_INSUFFICIENT_INFORMATION,
                sprintf('The session quote reference [%s] was not found.', $this->getParentQuoteId()),
                404
            );
        }
    }

    /**
     * Verifies that the immutable quote exists.
     *
     * @throws Bolt_Boltpay_BadInputException if immutable quote data is empty
     * @throws Varien_Exception if unable to get immutable quote
     */
    public function validateImmutableQuote()
    {
        $immutableQuote = $this->getImmutableQuote();

        if ($immutableQuote->isEmpty()) {
            $this->setErrorResponseAndThrowException(
                self::ERR_INSUFFICIENT_INFORMATION,
                sprintf(
                    'The immutable quote reference [%s] was not found.',
                    $this->boltHelper()->getImmutableQuoteIdFromTransaction($this->mockTransaction)
                ),
                404
            );
        }
    }

    /**
     * Verifies that the immutable quote is not empty and has items in the cart.
     *
     * @throws Exception
     */
    public function validateEmptyCart()
    {
        $immutableQuote = $this->getImmutableQuote();
        if (!$immutableQuote->getItemsCount()) {
            $this->setErrorResponseAndThrowException(
                self::ERR_INSUFFICIENT_INFORMATION,
                sprintf('The quote for order reference [%s] is empty.', $immutableQuote->getId()),
                422
            );
        }
    }

    /**
     * Verifies that the coupon isn't expired
     *
     * @throws Exception
     */
    public function validateToDateForRule()
    {
        $rule = $this->getRule();
        $date = $rule->getToDate();
        if ($date && date('Y-m-d', strtotime($date)) < date('Y-m-d')) {
            $this->setErrorResponseAndThrowException(
                self::ERR_CODE_EXPIRED,
                sprintf('The coupon code [%s] is expired.', $this->getCouponCode()),
                422
            );
        }
    }

    /**
     * Verifies that the from date is in the past.
     *
     * @throws Exception
     */
    public function validateFromDateForRule()
    {
        $rule = $this->getRule();
        $date = $rule->getFromDate();
        if ($date && date('Y-m-d', strtotime($date)) > date('Y-m-d')) {
            $desc = 'Code available from ' . date('m/d/Y', strtotime($date));
            $this->setErrorResponseAndThrowException(
                self::ERR_CODE_NOT_AVAILABLE,
                $desc,
                422
            );
        }
    }

    /**
     * Verifies that the coupon hasn't been used too many times.
     *
     * @throws Exception
     */
    public function validateCouponUsageLimits()
    {
        $coupon = $this->getCoupon();
        if ($coupon->getUsageLimit() && $coupon->getTimesUsed() >= $coupon->getUsageLimit()) {
            $this->setErrorResponseAndThrowException(
                self::ERR_CODE_LIMIT_REACHED,
                sprintf('The code [%s] has exceeded usage limit.', $this->getCouponCode()),
                422
            );
        }
    }

    /**
     * Verifies that the coupon hasn't been used too many times for the current customer.
     *
     * @throws Exception
     */
    public function validateCouponCustomerUsageLimits()
    {
        $coupon = $this->getCoupon();
        $immutableQuote = $this->getImmutableQuote();

        if ($customerId = $immutableQuote->getCustomerId()) {
            if ($customerId && $coupon->getUsagePerCustomer()) {
                $couponUsage = new Varien_Object();
                Mage::getResourceModel('salesrule/coupon_usage')->loadByCustomerCoupon(
                    $couponUsage,
                    $customerId,
                    $coupon->getId()
                );

                if ($couponUsage->getCouponId() && $couponUsage->getTimesUsed() >= $coupon->getUsagePerCustomer()) {
                    $this->setErrorResponseAndThrowException(
                        self::ERR_CODE_LIMIT_REACHED,
                        sprintf('The code [%s] has exceeded usage limit.', $this->getCouponCode()),
                        422
                    );
                }
            }
        }
    }

    /**
     * Verifies that the rule hasn't been used too many times for the current customer.
     *
     * @throws Exception
     */
    public function validateRuleCustomerUsageLimits()
    {
        $rule = $this->getRule();
        $immutableQuote = $this->getImmutableQuote();
        if ($usesPerCustomer = $rule->getUsesPerCustomer()) {
            $customerId = $immutableQuote->getCustomerId();
            $ruleCustomer = Mage::getModel('salesrule/rule_customer')->loadByCustomerRule($customerId, $rule->getId());

            if ($ruleCustomer->getId() && $ruleCustomer->getTimesUsed() >= $usesPerCustomer) {
                $this->setErrorResponseAndThrowException(
                    self::ERR_CODE_LIMIT_REACHED,
                    sprintf('The code [%s] has exceeded usage limit.', $this->getCouponCode()),
                    422
                );
            }
        }
    }

    /**
     * Attempts to apply the coupon to both the immutable and session quote.
     *
     * @throws Exception
     */
    protected function applyCouponToQuotes()
    {
        try {

            $parentQuote = $this->getParentQuote();
            $immutableQuote = $this->getImmutableQuote();
            $couponCode = $this->getCouponCode();

            // Try applying to parent first
            $this->applyCouponToQuote($parentQuote, $couponCode);
            // Apply coupon to clone
            $this->applyCouponToQuote($immutableQuote, $couponCode);
        } catch (\Exception $e) {
            $this->setErrorResponseAndThrowException(
                self::ERR_SERVICE,
                $e->getMessage(),
                422
            );
        }
    }

    /**
     * Applies the coupon code to the quote and recalculates totals.
     *
     * @param Mage_Sales_Model_Quote $quote
     * @param string $couponCode
     */
    protected function applyCouponToQuote($quote, $couponCode)
    {
        $quote->getShippingAddress()->setCollectShippingRates(true);
        $quote->setCouponCode($couponCode);

        $this->boltHelper()->collectTotals($quote, true)->save();
    }

    /**
     * Verifies that the coupon code was applied to the immutable quote.
     *
     * @throws Exception if coupon in quote doesn't match the coupon applied.
     */
    protected function validateAfterApplyingCoupon()
    {
        $immutableQuote = $this->getImmutableQuote();
        if ($immutableQuote->getCouponCode() != $this->getCouponCode()) {
            $this->setErrorResponseAndThrowException(
                self::ERR_SERVICE,
                sprintf("Invalid coupon code response for coupon [%s]", $this->getCouponCode()),
                422
            );
        }

        if (!$immutableQuote->validateMinimumAmount()) {
            $this->applyCouponToQuote($immutableQuote, '');
            $this->applyCouponToQuote($this->getParentQuote(), '');
            $this->setErrorResponseAndThrowException(
                self::ERR_MINIMUM_CART_AMOUNT_REQUIRED,
                $this->boltHelper()->__(
                    "The minimum order amount of %s has not been met",
                    $this->boltHelper()->getMinOrderAmountInStoreCurrency($immutableQuote->getStoreId())
                ),
                422
            );
        }
    }

    /**
     * Sets the response error and the http status code and then throws an exception.
     *
     * @param int $errCode Bolt error code
     * @param string $message of the error
     * @param int $httpStatusCode to be used in response
     * @param Exception|null $exception to be logged and thrown
     *
     * @throws Exception if $exception argument is provided
     * @throws Bolt_Boltpay_BadInputException if $exception argument is not provided
     */
    protected function setErrorResponseAndThrowException($errCode, $message, $httpStatusCode, \Exception $exception = null)
    {
        $this->responseError = array(
            'code'    => $errCode,
            'message' => $message,
        );

        $this->httpCode = $httpStatusCode;
        $this->responseCart = $this->getCartTotals();

        if ($exception){
            $this->boltHelper()->logException($exception);
            throw $exception;
        }
        $this->boltHelper()->logWarning($message);
        throw new \Bolt_Boltpay_BadInputException($message);
    }

    /**
     * Gets all cart totals from Bolt buildCart.
     *
     * @return array cart totals
     * @throws Varien_Exception if unable to retrieve immutable quote
     * @throws Mage_Core_Model_Store_Exception if unable to build cart
     */
    protected function getCartTotals()
    {
        $quote = $this->getImmutableQuote();
        /** @var Bolt_Boltpay_Model_BoltOrder $quoteModel */
        $boltOrderModel = Mage::getModel('boltpay/boltOrder');
        $cart = $boltOrderModel->buildCart($quote, true);

        return array(
            'total_amount' => $cart['total_amount'],
            'tax_amount'   => @$cart['tax_amount'],
            'discounts'    => $cart['discounts'],
        );
    }

    /**
     * Sets the success response (discount code, amount, description, and type). Also sets the HTTP status to 200.
     *
     * @return array
     * @throws Varien_Exception
     */
    protected function setSuccessResponse()
    {
        $this->responseCart = $this->getCartTotals();

        $immutableQuote = $this->getImmutableQuote();
        $address = $immutableQuote->isVirtual() ?
            $immutableQuote->getBillingAddress() :
            $immutableQuote->getShippingAddress();

        $this->discountInfo = array(
            'discount_code'   => $this->getCouponCode(),
            'discount_amount' => abs(round($address->getDiscountAmount() * 100)),
            'description'     => $this->boltHelper()->__('Discount ') . $address->getDiscountDescription(),
            'discount_type'   => $this->convertToBoltDiscountType($this->getRule()->getSimpleAction()),
        );

        $this->httpCode = 200;
    }

    /**
     * Maps the Magento discount type to a Bolt discount type
     *
     * @param $type
     *
     * @return string
     */
    protected function convertToBoltDiscountType($type)
    {
        switch ($type) {
            case "by_fixed":
            case "cart_fixed":
                return "fixed_amount";
            case "by_percent":
                return "percentage";
            case "by_shipping":
                return "shipping";
        }

        return "";
    }

    /**
     * Gets the session quote if it exists.
     *
     * @return Mage_Sales_Model_Quote|null
     * @throws Varien_Exception
     */
    protected function getParentQuote()
    {
        if (!$this->parentQuote) {
            if (!$this->requestObject) {
                return null;
            }

            /** @var Mage_Sales_Model_Quote $parentQuote */
            $parentQuote = Mage::getModel('sales/quote')->loadByIdWithoutStore($this->getParentQuoteId());
            $this->parentQuote = $parentQuote;
        }

        return $this->parentQuote;
    }

    /**
     * Gets the immutable quote if it exists.
     *
     * @return Mage_Sales_Model_Quote|null
     * @throws Varien_Exception
     */
    protected function getImmutableQuote()
    {
        if (!$this->immutableQuote) {
            if (!$this->requestObject) {
                return null;
            }

            /** @var Mage_Sales_Model_Quote $immutableQuote */
            $immutableQuote = Mage::getModel('sales/quote')->loadByIdWithoutStore(
                $this->boltHelper()->getImmutableQuoteIdFromTransaction($this->mockTransaction)
            );
            $this->immutableQuote = $immutableQuote;
        }

        return $this->immutableQuote;
    }

    /**
     * Get the coupon based on the coupon code if it exists.
     *
     * @return Mage_SalesRule_Model_Coupon|null
     */
    protected function getCoupon()
    {
        if (!$this->coupon) {
            if (!$this->requestObject) {
                return null;
            }
            /** @var Mage_SalesRule_Model_Coupon $coupon */
            $coupon = Mage::getModel('salesrule/coupon')->load($this->getCouponCode(), 'code');
            $this->coupon = $coupon;
        }

        return $this->coupon;
    }

    /**
     * Gets rule based on the coupon.
     *
     * @return Mage_SalesRule_Model_Rule|null
     */
    protected function getRule()
    {
        if (!$this->rule) {
            if (!$this->requestObject) {
                return null;
            }

            $coupon = $this->getCoupon();

            // Load the coupon discount rule
            /** @var Mage_SalesRule_Model_Rule $rule */
            $rule = Mage::getModel('salesrule/rule')->load($coupon->getRuleId());
            $this->rule = $rule;
        }

        return $this->rule;
    }

    /**
     * Gets the coupon code from the request.
     *
     * @return string
     */
    protected function getCouponCode()
    {
        return @$this->requestObject->discount_code;
    }

    /**
     * Gets session quote id from the request.
     *
     * @return int
     */
    public function getParentQuoteId()
    {
        return @$this->requestObject->cart->order_reference;
    }

    /**
     * Get response data for Bolt response
     *
     * @return array
     */
    public function getResponseData()
    {
        if ($this->responseError) {
            return array(
                'status' => 'error',
                'error'  => $this->responseError,
                'cart'   => $this->responseCart
            );
        }

        return array_merge(
            array(
                'status' => 'success',
                'cart'   => $this->responseCart
            ),
            $this->discountInfo
        );
    }

    /**
     * Gets HTTP code for Bolt response.
     *
     * @return int
     */
    public function getHttpCode()
    {
        return $this->httpCode;
    }

    /**
     * @param $code
     *
     * @return Mage_SalesRule_Model_Coupon
     * @throws Exception
     */
    public function decreaseCouponTimesUsed($code)
    {
        /** @var Mage_SalesRule_Model_Coupon $coupon */
        $coupon = Mage::getModel('salesrule/coupon')->load($code, 'code');

        $coupon->setTimesUsed($coupon->getTimesUsed() - 1);
        $coupon->save();
        return $coupon;
    }

    /**
     * @param                             $customerId
     * @param Mage_SalesRule_Model_Coupon $coupon
     */
    public function decreaseCustomerCouponTimesUsed($customerId, Mage_SalesRule_Model_Coupon $coupon)
    {
        $couponUsage = new Varien_Object();
        Mage::getResourceModel('salesrule/coupon_usage')->loadByCustomerCoupon($couponUsage, $customerId,
            $coupon->getId());

        if ($couponUsage->getCouponId()){
            /** @var Mage_Core_Model_Resource $resource */
            $resource = Mage::getSingleton('core/resource');
            /** @var Magento_Db_Adapter_Pdo_Mysql $writeConnection */
            $writeConnection = $resource->getConnection('core_write');
            $table = $resource->getTableName('salesrule/coupon_usage');
            $where = array('coupon_id = ?' => $coupon->getId(), 'customer_id = ?' => $customerId,);

            if ($couponUsage->getTimesUsed() > 1){
                $writeConnection->update($table, array('times_used' => $couponUsage->getTimesUsed() - 1), $where);
            } else {
                $writeConnection->delete($table, $where);
            }
        }
    }

    /**
     * @param                             $customerId
     * @param                             $ruleId
     *
     * @throws Exception
     */
    public function decreaseCustomerRuleTimesUsed($customerId, $ruleId)
    {
        /** @var Mage_SalesRule_Model_Rule_Customer $customerRule */
        $customerRule = Mage::getModel('salesrule/rule_customer')->loadByCustomerRule($customerId, $ruleId);

        if ($customerRule->getId()){
            $customerRule->setTimesUsed($customerRule->getTimesUsed() - 1);
            $customerRule->save();
        }
    }
}